/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * PALETTE.C
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 04/01/91
 *   MODIFIED : 10/08/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Palette manipulation routines.
 *
 * HISTORY
 *		10/08/94 (dcc) - updated to use <echidna/alloc.h>
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <stdlib.h>
#include <string.h>
#include <echidna/etypes.h>
#include <echidna/alloc.h>
#include <echidna/graphics.h>
#include <echidna/memory.h>
#include <echidna/eerrors.h>

#if !USE_GRAFX

/**************************** C O N S T A N T S ***************************/


/******************************** T Y P E S *******************************/


/****************************** G L O B A L S *****************************/

short PaletteMaxColors;
volatile short PaletteIsFading;
volatile short CycleSetIsFading;
short ColorCyclingActive;

/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/


/*********************************************************************
 *
 * SizeOfPalette
 *
 * SYNOPSIS
 *		size_t SizeOfPalette (PaletteType *palette)
 *
 * PURPOSE
 *		Return space needed to make a copy of the palette.
 *
 * INPUT
 *		palette: Pointer to palette structure.
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Returns number of bytes required to make copy of palette.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
size_t SizeOfPalette (PaletteType *palette)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "SizeOfPalette";
#endif
#if REQUIRE
	if (!palette) {
		SetGlobalErr ( ERR_GENERIC);
		GEprintf ("SizeOfPalette:Null passed for palette");
		return 0;
	}
#endif
/*-------------------------------------------------------------------------*/
{
#if __MSDOS__
	size_t color_size;

	switch (GraphicsMode) {
	case MODE_EGA: 
		color_size = sizeof (UBYTE);
		break;
	case MODE_MCGA:
		color_size = sizeof (ColorType);
		break;
	default:
		/* MODE_CGA */
		color_size = 0;
		break;
	}
	return sizeof (PaletteType) + palette->NumColors * color_size;

#elif __AMIGAOS__
	return sizeof (PaletteType) + palette->NumColors * sizeof (ColorType);
#else
#error SizOfPalette () not implemented for this platform.
#endif

}
} /* SizeOfPalette */


/*********************************************************************
 *
 * CopyPalette
 *
 * SYNOPSIS
 *		PaletteType *CopyPalette (
 *			void *dstbuf,
 *			PaletteType *srcpal
 *		)
 *
 * PURPOSE
 *		To copy a palette into already allocated space.
 *
 * INPUT
 *		dstbuf: Buffer at least as big as value returned by call to
 *					SizeOfPalette (srcpal).
 *		srcpal: Pointer to palette.
 *
 * EFFECTS
 *		Makes copy of palette pointed to by 'srcpal' in space pointed
 *		to by 'dstbuf'.
 *
 * RETURN VALUE
 *		On success, returns (PaletteType *)dstbuf, after having created a 
 *		copy of	the palette in that space.
 *		On Failure returns NULL.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		SizeOfPalette().
 *	
*/
PaletteType *CopyPalette (
	void *dstbuf,
	PaletteType *srcpal
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CopyPalette";
#endif
#if REQUIRE
	if (!dstbuf) {
		SetGlobalErr ( ERR_GENERIC);
		GEprintf ("CopyPalette: dstbuf is Null");
		return NULL;
	}
	if (!srcpal) {
		SetGlobalErr ( ERR_GENERIC);
		GEprintf ("CopyPalette: srcpal is Null");
		return NULL;
	}
#endif
/*-------------------------------------------------------------------------*/
{

#if __MSDOS__
	size_t color_size;
	void *color_mem;
	BYTE *colordata = (BYTE *)dstbuf + sizeof (PaletteType);

	*((PaletteType *)dstbuf) = *srcpal;
	switch (GraphicsMode) {
	case MODE_EGA: 
		color_mem = srcpal->Data.EGA;
		color_size = sizeof (UBYTE);
		break;
	case MODE_MCGA:
		color_mem = srcpal->Data.RGB;
		color_size = sizeof (ColorType);
		break;
	default:
		/* MODE_CGA */
		return (PaletteType *)dstbuf;
	}
	memcpy (colordata, 
			color_mem, 
			color_size * srcpal->NumColors);

	switch (GraphicsMode) {
	case MODE_EGA: 
		((PaletteType *)dstbuf)->Data.EGA = (UBYTE *)colordata;
		break;
	case MODE_MCGA:
		((PaletteType *)dstbuf)->Data.RGB = (ColorType *)colordata;
		break;
	}
#elif __AMIGAOS__
	BYTE *colordata = (BYTE *)dstbuf + sizeof (PaletteType);

	*((PaletteType *)dstbuf) = *srcpal;
	memcpy (colordata,
			srcpal->Data.RGB, 
			sizeof (ColorType) * srcpal->NumColors);

	((PaletteType *)dstbuf)->Data.RGB = (ColorType *)colordata;
	
#else
#error SizOfPalette () not implemented for this platform.
#endif

	return (PaletteType *)dstbuf;
}
} /* CopyPalette */

/*********************************************************************
 *
 * DupPalette
 *
 * SYNOPSIS
 *		PaletteType *DupPalette (PaletteType *palette)
 *
 * PURPOSE
 *		Get allocated copy of palette.
 *
 * INPUT
 *		palette: pointer to palette.
 *
 * EFFECTS
 *		Allocates space for copy.
 *
 * RETURN VALUE
 *		On success, returns pointer to allocated copy.
 *		On failure, returns NULL.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
PaletteType *DupPalette (PaletteType *palette)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DupPalette";
#endif
#if REQUIRE
	if (!palette) {
		SetGlobalErr ( ERR_GENERIC);
		GEprintf ("DupPalette: palette is Null");
		return NULL;
	}
#endif
/*-------------------------------------------------------------------------*/
{
	BYTE *buf;

	/** Allocate ColorCycleSetType structure **/
	if ((buf = (BYTE *)MEM_calloc (1, SizeOfPalette (palette))) == NULL)
	{
		SetGlobalErr ( ERR_OUT_OF_MEMORY);
		GEprintf ("DupPalette:OOM allocating palette mem");
		
	} else {
		CopyPalette (buf, palette);
	}
	
	return (PaletteType *)buf;
}
} /* DupPalette */


/*********************************************************************
 *
 * WaitForFade
 *
 * SYNOPSIS
 *		void WaitForFade (void) 
 *
 * PURPOSE
 *		To wait in loop until fading is completed.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void WaitForFade (void) 
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "WaitForFade";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{

	/* Wait till interrupt signals it is done. */
	while (PaletteIsFading || CycleSetIsFading) {}
}
} /* WaitForFade */

#endif /* !USE_GRAFX */

