/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * LISTFUNC.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 08/01/89
 *   MODIFIED : 10/10/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w-----
 *	 E C H i D N A
 *	--------------
 *
 * DESCRIPTION
 *	 Functions for dealing with linked lists.
 *
 * HISTORY
 *		05/13/93 (dcc) - added InsertList().
 *		10/05/94 (dcc) - updated to use <echidna/alloc.h>
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <stdlib.h>
#include <echidna/strings.h>
#include <echidna/listfunc.h>
#include <echidna/eerrors.h>
#include <echidna/alloc.h>

#if __NOT_ANSI__
#define SIZETYPE(val)	val
#else
#define SIZETYPE(val)	(size_t)(val)
#endif

#define ND(node)	((NodeType *)(node))
#undef Next
#undef Prev
#undef CreateNode

#define Next(node)		((void *)(((NodeType *)(node))->Next))
#define Prev(node)		((void *)(((NodeType *)(node))->Prev))

/* Use the following macros when the expression will be a lvalue */

#define NextL(node)		(((NodeType *)(node))->Next)
#define PrevL(node)		(((NodeType *)(node))->Prev)

                /* 0123456789012345 */
#define SIGNATURE "A Valid NodeType"

#if DEBUG_LISTFUNC
extern void checknode (void *nd);
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_checknode                                            */
/*------------------------------------------------------------------------*/

#if DEBUG_LISTFUNC
#include <stdio.h>
void checknode (void *nd)
{
	if (!IsEOList (nd) && !IsSOList (nd) && ND(nd)->Type == 1) {
		if (memcmp (SIGNATURE, NodeName(nd), 16)) {
			puts ("Bad Node!\n");
		}
	}
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_InsertBefore                                         */
/*------------------------------------------------------------------------*/

void InsertBefore (oldnode, newnode)
	void *oldnode;
	void *newnode;
{
	void *temp;

	#if DEBUG_LISTFUNC
	checknode (oldnode);
	checknode (newnode);
	#endif
	temp           = Prev(oldnode);
	PrevL(newnode) = temp;
	NextL(newnode) = oldnode;
	NextL(temp)    = newnode;
	PrevL(oldnode) = newnode;
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_InsertAfter                                          */
/*------------------------------------------------------------------------*/

void InsertAfter (oldnode, newnode)
	void *oldnode;
	void *newnode;
{
	void *temp;

	#if DEBUG_LISTFUNC
	checknode (oldnode);
	checknode (newnode);
	#endif
	temp           = Next(oldnode);
	PrevL(newnode) = oldnode;
	NextL(newnode) = temp;
	PrevL(temp)    = newnode;
	NextL(oldnode) = newnode;
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_Remove                                               */
/*------------------------------------------------------------------------*/

#if __AMIGAOS__
#else
void Remove (node)
	void *node;
{
	NodeType *prev;
	NodeType *next;

	#if DEBUG_LISTFUNC
	checknode (node);
	#endif
	prev = Prev(node);
	next = Next(node);

	NextL(prev) = next;
	PrevL(next) = prev;
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_AddHead                                              */
/*------------------------------------------------------------------------*/

#if __AMIGAOS__
#else
void AddHead (list, node)
	ListType *list;
	void	 *node;
{
	InsertAfter ((NodeType *)list, node);
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_RemHead                                              */
/*------------------------------------------------------------------------*/

#if __AMIGAOS__
#else
void *RemHead (list)
	ListType *list;
{
	NodeType *node;

	node = Head(list);
	if (!IsEOList(node)) {
		Remove (node);
		#if DEBUG_LISTFUNC
		checknode (node);
		#endif
	} else {
		node = NULL;
	}
	return (void *)node;
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_AddTail                                              */
/*------------------------------------------------------------------------*/

#if __AMIGAOS__
#else
void AddTail (list, node)
	ListType *list;
	void	 *node;
{
	InsertBefore ((NodeType *)&list->Tail, node);
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_RemTail                                              */
/*------------------------------------------------------------------------*/

#if __AMIGAOS__
#else
void *RemTail (list)
	ListType *list;
{
	NodeType *node;

	node = Tail(list);
	if (!IsSOList(node)) {
		Remove (node);
		#if DEBUG_LISTFUNC
		checknode (node);
		#endif
	} else {
		node = NULL;
	}
	return (void *)node;
}

#endif

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_InitList                                             */
/*------------------------------------------------------------------------*/

void InitList (list)
	ListType *list;
{
	list->Head	= (NodeType *)&list->Tail;
	list->Tail	= NULL;
	list->TailPred	= (NodeType *)&list->Head;
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_CreateList                                           */
/*------------------------------------------------------------------------*/

ListType *CreateList(name)
	char *name;
{
	ListType *list;

	list = MEM_malloc (sizeof (ListType));
	if (list) {
		clearstruct ((char *)list, (int)sizeof(ListType));
		InitList (list);
		if (name) {
			if (!AddName ((NodeType *)&list->Tail, name)) {
				list = DeleteList (list);
			}
		}
	}
	return (list);
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_EmptyList                                            */
/*------------------------------------------------------------------------*/

void EmptyList (list)
	ListType *list;
{
	NodeType	*node;

	if (list) {
		while ((node = RemTail (list)) != NULL) {
			DeleteNode (node);
		}
	}
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_DeleteList                                           */
/*------------------------------------------------------------------------*/

ListType *DeleteList(list)
	ListType *list;
{
	if (list) {
		EmptyList (list);
		FreeName ((NodeType *)&list->Tail);
		MEM_free (list);
	}

	return (NULL);
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_AddName                                              */
/*------------------------------------------------------------------------*/

char *AddName(node, name)
	void		*node;
	char		*name;
{
	register char	*s;
	register int	size;

	size = strlen(name) + 1;
	s    = MEM_malloc(SIZETYPE(size));
	if (s) {
		strcpy (s, name);
	}
	ND(node)->Name = s;

	return s;
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_FreeName                                             */
/*------------------------------------------------------------------------*/

void FreeName (node)
	void *node;
{
	register char	*name = ND(node)->Name;

	#if DEBUG_LISTFUNC
	checknode (node);
	#endif
	if (name) {
		MEM_free (name);
	}
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_Find_A_Name                                          */
/*------------------------------------------------------------------------*/

typedef int (*StrFunc)(char *s1, char *s2);
void	*Find_A_Name (
	void	*node,
	char	*str,
	short	 insensitive
)
{
	StrFunc strfunc;

	strfunc = (insensitive) ? (StrFunc)stricmp : (StrFunc)strcmp;
	while (! IsEOList(node)) {
		if (! IsSOList(node)) {
			if (! strfunc (((NodeType *)node)->Name, str)) {
				return node;
			}
		}
		#if DEBUG_LISTFUNC
		checknode (node);
		#endif
		node = Next (node);
	}

	return NULL;
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_CreateNode                                           */
/*------------------------------------------------------------------------*/

void *CreateNode(size, name)
	size_t	 size;
	char	*name;
{
	register NodeType	*node;

	node = MEM_malloc (SIZETYPE(size));
	if (node) {
		clearstruct ((char *)node, SIZETYPE(size));
		if (name) {
			if (!AddName (node, name)) {
				node = DeleteNode (node);
			}
		}
	}
	return (void *)node;
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_CreateNode2                                          */
/*------------------------------------------------------------------------*/

#if DEBUG_LISTFUNC
void *CreateNode2(size, name)
	size_t	 size;
	char	*name;
{
	register NodeType	*node;

	node = MEM_malloc (SIZETYPE(size));
	if (node) {
		clearstruct ((char *)node, SIZETYPE(size));
		if (!name) {
			name = SIGNATURE;
			node->Type = 1;
		}
		if (name) {
			if (!AddName (node, name)) {
				node = DeleteNode (node);
			}
		}
	}
	return (void *)node;
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_DeleteNode                                           */
/*------------------------------------------------------------------------*/

/*
 * Delete Node from List
 *
 *		10/10/94 (dcc) - make sure <node> != NULL b4 calling checknode()
 *
 * return "Next" node in list.
 */

void *DeleteNode(node)
	void *node;
{
	if (node) {
		#if DEBUG_LISTFUNC
		checknode (node);
		#endif
		FreeName (node);
		MEM_free(node);
	}
	return NULL;
}

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_clearstruct                                          */
/*------------------------------------------------------------------------*/

#ifndef clearstruct
void clearstruct (mem, size)
	char		*mem;
	int		size;
{
	for (;size > 0; --size) {
		*mem++ = '\0';
	}
}
#endif

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_OpOnList                                             */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * OpOnList
 *
 * SYNOPSIS
 *		long OpOnList (void *node, OpOnListFunc func)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long OpOnList (void *node, OpOnListFunc func)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "OpOnList";
#endif
#if REQUIRE
	if (!node) {
		SetGlobalErr (ERR_GENERIC);
		GEprintf ("Passed NULL to OpOnList()");
		return (-1);
	}
#endif
/*-------------------------------------------------------------------------*/

	{
		long	count = 0;

		while (!IsEOList (node)) {
			if (!IsSOList (node)) {
				count++;
				#if DEBUG_LISTFUNC
				checknode (node);
				#endif
				if (func) {
					node = func (node);
					if (!node) {
						return (-1);
					}
				} else {
					node = Next (node);
				}
			} else {
				node = Next (node);
			}
		}
		return count;
	}

} /* OpOnList */

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_InsertInOrder                                        */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * InsertInOrder
 *
 * SYNOPSIS
 *		void	*InsertInOrder (
 *			void	*inode,
 *			void	*node
 *		)
 *
 * PURPOSE
 *		To insert a list node in alphabetical order by Node.Name field.
 *
 * INPUT
 *		inode:	node of list to be inserted into.
 *		node:	node to insert into list.
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Returns node inserted.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void	*InsertInOrder (
	void		*inode,
	void		*node
)
{
	while (! IsEOList (inode)) {
		if (!IsSOList (inode)) {
			if (strcmp (ND(inode)->Name, (ND(node)->Name)) > 0) {
				break;
			}
		}
		inode = Next (inode);
	}
	
	InsertBefore (inode, node);

	return node;
} /* InsertInOrder */

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_InsertList                                           */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * InsertList
 *
 * SYNOPSIS
 *		ListType *InsertList(NodeType *pnod1, ListType *pnod2)
 *
 * PURPOSE
 *		Insert plst2 and all nodes linked after plst2 into list
 *		before pnod1.
 *
 * USAGE
 *		plstNew = InsertLists(pnod1, plst2);
 *
 * INPUT
 *		Both pnod1 & pnod2 must be non-null.
 *		Either or both may be empty lists.
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Returns pnod1.
 *
 * HISTORY
 *		05/11/93 Tuesday (dcc) - created.
 *
 * SEE ALSO
 *
*/
void *InsertList(void *pnod1, void *pnod2)
{
	NodeType *pnodNext;

	if (IsEOList(pnod2))
		return pnod1;

	pnodNext = Next(pnod2);
	Remove(pnod2);
	InsertBefore(pnod1, pnod2);

	if (IsEOList(pnodNext))
		return pnod2;

	return MergeLists(pnod2, pnodNext);
} /* InsertList */

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_MergeLists                                           */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * MergeLists
 *
 * SYNOPSIS
 *		ListType	*MergeLists (NodeType *node1, NodeType *node2);
 *
 * PURPOSE
 *		To Merge-enate two lists into one.  Moves all nodes of from Node2
 *		to end of list to just after node1.
 *
 * INPUTS
 *		Node1 = pointer to a node in a list
 *		Node2 = pointer to a node in a list
 *
 * RESULTS
 *		Returns pointer to a node1 with all nodes after and including
 *		node2 in their original order.
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void	*MergeLists (
	void	*node1,
	void	*node2
)
{
	NodeType	*nextnode;
	NodeType	*lastnode;

	lastnode = node1;
	while (!IsEOList (node2)) {
		nextnode = Next (node2);
		Remove (node2);
		InsertAfter (lastnode, node2);
		lastnode = node2;
		node2    = nextnode;
	}

	return node1;
} /* MergeLists */

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_ConcatLists                                          */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ConcatLists
 *
 * SYNOPSIS
 *		ListType	*ConcatLists (ListType *list1, ListType *list2);
 *
 * USAGE
 *		NewList = ConcatLists (list1, list2);
 *
 * PURPOSE
 *		To concatenate two lists into one.
 *
 * INPUTS
 *		Both list1 and list2 must be non-null.  
 *		Either or both may be empty lists.
 *
 * RESULTS
 *		Returns pointer to a list with the elements of 'list1' followed by
 *		the elements of 'list2' in their original order.
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
ListType	*ConcatLists (
	ListType	*list1,
	ListType	*list2
)
{
	MergeLists (Tail (list1), Head (list2));

	return list1;
} /* ConcatLists */

/*------------------------------------------------------------------------*/
/**# MODULE:LISTFUNC_NextPrev                                             */
/*------------------------------------------------------------------------*/

/*
** This module must be last
*/

#if DEBUG_LISTFUNC
#undef Next
#undef Prev

/*		10/08/94 (dcc) - make sure return <nd> != NULL b4 calling checknode() */

void *Next (void *nd)
{
	checknode (nd);
	nd = ND(nd)->Next;
	if (nd)
		checknode (nd);
	return nd;
}
void *Prev (void *nd)
{
	checknode (nd);
	nd = ND(nd)->Prev;
	if (nd)
		checknode (nd);
	return nd;
}
#endif

