/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * INTERUPT.C
 *
 * PROGRAMMER : Dan Chang
 *    VERSION : 00.000
 *    CREATED : 04/10/91
 *   MODIFIED : 07/27/94
 *       TABS : 04 07
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Support for interrupts and timers.
 *
 * HISTORY
 *		04/10/91..04/18/91 (dcc) - Initial version.
 *		09/13/93 Monday (dcc) - MS-DOS protected mode support added.
 *		12/04/93 Saturday (dcc) - allow multiple calls to CloseTimer().
 *
*/
#include <echidna/platform.h>
#include "switches.h"

#if __MSDOS__
#if __MSDOS32X__
#include <dos.h>
#include <string.h>
#else
#include <mem.h>
#endif/*__MSDOS32X__*/
#endif/*__MSDOS__*/

#include <echidna/etypes.h>

#include <echidna/timer.h>
#include "vblank.h"

/**************************** C O N S T A N T S ***************************/

/* How many times the timer counts in one second. */
/* DCC - this is not real number - look it up in some PC book someday... */
/* dcc - looked it up in one source, verify someday... */

#define TICKSPERSECOND	(1193180UL)	/* (1191400UL) */

/******************************** T Y P E S *******************************/


/****************************** G L O B A L S *****************************/


UWORD uwTimerTail;	/* offset of last timer record in use */

/******************************* M A C R O S ******************************/

#define Prev(Node) (((TimerType far32 *) ((BYTE far32 *) _TimerRecord+Node))->uwPrev)
#define Next(Node) (((TimerType far32 *) ((BYTE far32 *) _TimerRecord+Node))->uwNext)

/***************************** R O U T I N E S ****************************/

/* InitInterruptLinkList - reset the linked list of interrupt routine entries.
	The zero'th record is the head of the active list. It is never processed. */

static void InitInterruptLinkList(void)
{
#if DONT_FAKEIT
	UWORD i;

	uwTimerTail = sizeof(TimerType);		/* offset of end of list */
	_TimerRecord[0].uwPrev = 0xffff;		/* head of list marker */
	_TimerRecord[0].uwNext = sizeof(TimerType);		/* -> end of list marker */
	_TimerRecord[1].uwNext = 0xffff;		/* end of list marker */
	_TimerRecord[1].uwPrev = 0;				/* -> to beginning of list */

	for (i = 2; i < MAXTIMERRECORDS-1; i++)
	{
		_TimerRecord[i].uwNext = (UWORD) ((i+1) * sizeof(TimerType));
		_TimerRecord[i].uwPrev = (UWORD) ((i-1) * sizeof(TimerType));
	}
	_uwFreeHead = 2 * sizeof(TimerType);	/* offset of 1st free list */
	_TimerRecord[MAXTIMERRECORDS-1].uwNext = 0xffff;		/* end of free list marker */
	_TimerRecord[2].uwPrev = 0xffff;		/* end of free list marker */
#endif
}

/* AddInterruptRoutine - low level routine to add an interrupt to the
	timer record structure. Returns handle (array offset) of new
	timer record if successful, 0 if list is full. */

static WORD AddInterruptRoutine(void far32 *Address, ULONG ulTriggerCount, UBYTE ubType)
{
#if DONT_FAKEIT
	register UWORD uwNewNode;

	uwNewNode = _uwFreeHead;

	if (uwNewNode == 0xffff)
		return 0;

	/* Is new timer faster than current interrupt rate? */

	if (ulTriggerCount < (ULONG) _TimerCount)
		_NewTimerCount = (UWORD) ulTriggerCount;	/* YES, speed up interrupt rate */

	((TimerType far32 *) ((BYTE far32 *) _TimerRecord+uwNewNode))->ulAddress = (ULONG) Address;
	((TimerType far32 *) ((BYTE far32 *) _TimerRecord+uwNewNode))->ulCounter = 0;
	((TimerType far32 *) ((BYTE far32 *) _TimerRecord+uwNewNode))->ulTrigger =
				(ulTriggerCount & 0xffffffUL) | ((ULONG) ubType << 24);

	/* Update free list 1st entry offset. */

	_uwFreeHead = Next(uwNewNode);

	/* Link in to active list in front of tail of active list.
		Prev = Prev(uwTimerTail), Next = uwTimerTail. */

	Prev(uwNewNode) = Prev(uwTimerTail);
	Next(uwNewNode) = uwTimerTail;
	Prev(uwTimerTail) = uwNewNode;

	/* Add new record to end of active timer list. */
	/* Note that record is completely correct before we add to end of list. */

	Next(Prev(uwNewNode)) = uwNewNode;

	/* Return active timer list (truncate to WORD). */

	return (WORD) uwNewNode;
#endif
}

/*
	bits   1 or 0
	7 - ACTIVE or not
	6 - record being MODIFYed or not
	3 - UP or DOWN timer
	2 - ROUTINE or TIMER
	1 & 0 - size to increment (0 = byte, 1 = word, 3 = long
*/

#define ACTIVE		(1 << 7)
#define MODIFY		(1 << 6)
#define ROUTINE	(1 << 2)
#define UPTIMER	(1 << 3)
#define DOWNTIMER (0 << 3)

/* InstallInterrupt - install an interrupt routine. */

WORD far32 InstallInterrupt(InterruptRoutineType (*routine)(void), WORD rate)
{
#if DONT_FAKEIT
	ULONG ulCount;

	if (rate)
		ulCount = TICKSPERSECOND / rate;
	else
		ulCount = 0UL;

	return AddInterruptRoutine((void far32 *) routine, ulCount, ACTIVE | ROUTINE);
#endif
}

/* InstallTimer - install a timer. */

WORD far32 InstallTimer(void far32 *timer, WORD rate, WORD size, WORD direction)
{
#if DONT_FAKEIT
	ULONG ulCount;
	UBYTE ubType;

	if (rate)
		ulCount = TICKSPERSECOND / rate;
	else
		ulCount = 0UL;

	/* Set ubType to 0, 1, or 3 for byte, word, or long. */

	ubType = size-1;
	if (ubType > 3)				/* Bigger than four bytes? Set to long. */
		ubType = 3;
	if (ubType == 2)				/* Three bytes? Set to WORD. */
		ubType = 1;

	ubType |= (direction > 0) ? UPTIMER : DOWNTIMER;
	ubType |= ACTIVE;

	return AddInterruptRoutine((void far32 *) timer, ulCount, ubType);
#else
	return (1);
#endif
}

/* ChangeInterruptRate - change the rate an interrupt occurs at. */

BOOL far32 ChangeInterruptRate(WORD handle, WORD rate)
{
#if DONT_FAKEIT
	if (((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ubFlags & ACTIVE)
	{
		ULONG ulCount;
		UBYTE ubType;

		if (rate)
			ulCount = TICKSPERSECOND / rate;
		else
			ulCount = 0UL;

		/* Is new timer faster than current interrupt rate? */

		if (ulCount < (ULONG) _TimerCount)
			_NewTimerCount = (UWORD) ulCount;	/* YES, speed up interrupt rate */

		/* Lock record: set MODIFY bit in ubFlags. */

		ubType = (((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ubFlags |= MODIFY);

		/* Reset Counter and update Trigger value. */

		((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ulCounter = 0;
		((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ulTrigger =
					(ulCount & 0xffffffUL) | ((ULONG) ubType << 24);

		/* Remove lock. */

		((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ubFlags = ubType & ~MODIFY;

		return TRUE;
	}
	return FALSE;
#endif
}

/* ResetTimer - reset a timer count value. */

BOOL far32 ResetTimer(WORD handle, ULONG value)
{
#if DONT_FAKEIT
	if (((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ubFlags & ACTIVE)
	{
		UBYTE ubType;

		ubType = ((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ubFlags;

		if (ubType & ROUTINE)				/* Make sure it's not a routine! */
			return FALSE;	/* Sheesh. Some cretin is trying to reset a routine! */

		/* Lock record: set MODIFY bit in ubFlags. */

		((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ubFlags = ubType | MODIFY;

		/* Reset internal Counter and update value. */

		((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ulCounter = 0;
		if ((ubType & 3) == 0)
		{
			(*(UBYTE far32 *) ((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ulAddress) = (UBYTE) value;
		}
		else if ((ubType & 3) == 1)
		{
			(*(UWORD far32 *) ((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ulAddress) = (UWORD) value;
		}
		else
		{
			(*(ULONG far32 *) ((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ulAddress) = value;
		}

		/* Remove lock. */

		((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ubFlags = ubType;

		return TRUE;
	}
	return FALSE;
#endif
}

/* RemoveInterrupt - remove the interrupt so it will no longer be called. */

BOOL far32 RemoveInterrupt(WORD handle)
{
#if DONT_FAKEIT
	/* Just mark it so it is no longer active. */

	if (((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ubFlags & ACTIVE)
	{
		((TimerType far32 *) ((BYTE far32 *) _TimerRecord+handle))->ubFlags &= (UBYTE) ~ACTIVE;
#if 0
		/* Unlink entry from active list. */

		Next(Prev(handle)) = Next(handle);
		Prev(Next(handle)) = Prev(handle);

		/* Add node to head of free list. */

		Prev(_uwFreeHead) = handle;
		Next(handle) = _uwFreeHead;
		Prev(handle) = 0xffff;
		_uwFreeHead = (UWORD) handle;
#endif
		return TRUE;
	}
	return FALSE;
#endif
}

/* InitVBlanks - initialize array of active vertical blank routines. */

static void InitVBlanks(void)
{
#if DONT_FAKEIT
	_LastVBlankActive = 0;
	memset(_VBlankActive, FALSE, sizeof(_VBlankActive));
#endif
}

/* InstallVblank - install an interrupt routine that will be called
	during vertical blanking. Returns handle if successful, else 0. */

WORD far32 InstallVBlank(InterruptRoutineType (*routine)(void))
{
#if DONT_FAKEIT
	WORD i;

	for (i = 0; i < MAXVBLANKS; i++)
	{
		if (!_VBlankActive[i])
		{
			_VBlankAddress[i] = routine;
			_VBlankActive[i] = TRUE;

			if (++i > _LastVBlankActive)
				_LastVBlankActive = i;
			return i;
		}
	}
	return 0;
#endif
}

/* RemoveVBlank - remove an interrupt routine so it will no longer
	be called during vertical blanking. */

BOOL far32 RemoveVBlank(WORD handle)
{
#if DONT_FAKEIT
	if (_VBlankActive[--handle])
	{
		_VBlankActive[handle] = FALSE;
		while (_LastVBlankActive && !_VBlankActive[_LastVBlankActive-1])
			_LastVBlankActive--;
		return TRUE;
	}
	return FALSE;
#endif
}

/* OpenTimers - Opens timer routines, initializes internal variables. */
/* jma added check for already opened timers.*/

static int AlreadyOpenedTimers = FALSE;

void far32 OpenTimers(void)
{
#if DONT_FAKEIT
	if (AlreadyOpenedTimers)
		return;
	AlreadyOpenedTimers = TRUE;

	InitInterruptLinkList();
	InitVBlanks();
	_OpenTimerInterrupt();
#endif
}

/* CloseTimers - Restore system resources taken over by timer routines
	to original state. */

void far32 CloseTimers(void)
{
#if DONT_FAKEIT
	if (!AlreadyOpenedTimers)
		return;
	_CloseTimerInterrupt();
	AlreadyOpenedTimers = FALSE;
#endif
}

