/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * tUMEsup.c
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 11/06/89
 *   MODIFIED : 10/08/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w-----
 *	 E C H I D N A
 *	--------------
 *
 * DESCRIPTION
 *		Routines to load tables (offset and recovery) and images from
 *		tumepack created all images file.
 *		Images are converted upon loading to current graphics mode format.
 *
 * HISTORY
 *		11/30/89 (RGM) - Re-did LoadGameMap to ExtendedLoadGameMap, added
 *	                     macro in LodImijz.h for LoadGameMap() and
 *	                     ReLoadGameMap().
 *		12/17/89 (JMA) - Revised for new naming/error standards.
 *		12/18/89 (RGM) - Added support for 'conversion rooms', took
 *	                     out stuff for ExtendedLoadGameMap (now supported
 *	                     using new naming conventions).
 *		12/20/89 (GAT) - Changed DisplayRoom() to DrawRoom() with
 *						 Simpler parameters. Separated tUME support
 *						 from BrushPak support.
 *		12/21/89 (RGM) - Changed to true non-homogeneous conversion rooms.
 *		01/15/90 (GAT) - Added DrawPartOfRoom().
 *		10/07/94 (dcc) - updated to use <echidna/alloc.h>
 *
*/
#include <echidna/platform.h>
#include "switches.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <limits.h>
#include <echidna/etypes.h>
#include <echidna/alloc.h>
#include <echidna/emacros.h>
#include <echidna/rgmextra.h>			/* added */
#include <echidna/memory.h>
#include <echidna/eerrors.h>
#include <echidna/tumesup.h> 
#include <echidna/fpfload.h>
#include <echidna/graphics.h>
#include <echidna/pixlscan.h>
#include "xpaksup.h"
#include "tpmasks.h"

#if __MSDOS__
#define NUM_CGA_COLORS	4
#define READMODE	O_RDONLY|O_BINARY
#else
#define READMODE	O_RDONLY
#endif

/***************************** T Y P E D E F S ****************************/
typedef struct {
	UWORD	 NumTileSets;
	UWORD	 NumGlobalTileSets;
	UWORD 	 NumTiles;
	UWORD 	 NumGlobalTiles;
	UWORD	 RecoverTableSize;
	UWORD	 OffsetTableSize;
	UWORD 	 NumEGAPatmapTables;
	UWORD 	 NumCGAPatmapTables;
} CTIHeaderType;

typedef struct {
	WORD	RoomWidth;
	WORD	RoomHeight;
	WORD	NumLayers;
	WORD	TileWidth;
	WORD	TileHeight;
	UWORD 	 NumMCGAColors;
	UWORD 	 NumEGAColors;
	UWORD 	 CGAPaletteCode;
} TPRoomHeaderType;

typedef struct {
	WORD	RoomWidth;
	WORD	RoomHeight;
	WORD	NumLayers;
	WORD	TileWidth;
	WORD	TileHeight;
} TPDimHeaderType;

/****************************** G L O B A L S *****************************/

CTIType	*GlobalCTIPtr;

/******************************* M A C R O S ******************************/
#if __iAPx86__
#define makeid(d,c,b,a)   (((LONG)(a)<<24)|((LONG)(b)<<16)|((c)<<8)|(d))
#else
#define makeid(a,b,c,d)   (((LONG)(a)<<24)|((LONG)(b)<<16)|((c)<<8)|(d))
#endif

#define CHUNK_HEAD	makeid ('H','E','A','D')
#define CHUNK_RTAB	makeid ('R','T','A','B')
#define CHUNK_OTAB	makeid ('O','T','A','B')
#define CHUNK_EPRM	makeid ('E','P','R','M')
#define CHUNK_CPRM	makeid ('C','P','R','M')
#define CHUNK_TSAT	makeid ('T','S','A','T')
#define CHUNK_ISIZ	makeid ('I','S','I','Z')
#define CHUNK_IDAT	makeid ('I','D','A','T')

#define CHUNK_MPAL	makeid ('M','P','A','L')
#define CHUNK_EPAL	makeid ('E','P','A','L')
#define CHUNK_LDAT	makeid ('L','D','A','T')

#define CHUNK_NAME	makeid ('N','A','M','E')

/*********************************************************************
 *
 * ReadInTiles
 *
 * SYNOPSIS
 *		short	ReadInTiles (
 *			UWORD	 numtiles,
 *			CTIType	*cti,
 *			ULONG	*AttribTable,
 *			ULONG	*tilesizes
 *		)
 *
 * PURPOSE
 *		Read in One tile at a time and convert it to graphics format.
 *
 * INPUT
 *		numtiles: 		How many tiles to read.
 *		cti:			Partially fille out cti stucture.
 *		AttribTable:	Table of tileset attributes flags.
 *		tilesizes:		Table of tile sizes.
 *	
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Returns TRUE if succesful, else returns FALSE.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short	ReadInTiles (
	UWORD	 numtiles,
	CTIType	*cti,
	ULONG	*AttribTable,
	ULONG	*tilesizes
)
{
	short				 i;
	short	 			 type;
	ULONG	 			 size;
	ULONG				 attrib = 0;
	short				 tileset;
	short				 tilecount;
	short				 result = TRUE;
	GShape				*newshapeptr;
	ConvertSType		 cs;
#if __MSDOS__
	short	patindex, prevpatindex = -1;
#endif

	if (! OpenPixlScan (&cs, MAX_COLORS)) {
		return FALSE;
	}

	cs.HFlip = FALSE;
	cs.VFlip = FALSE;
	SetRemapFlag (cs,FALSE);

	/** Load and process Tiles **/
	tileset		= 0;
	tilecount	= 0;
	for (i = 0; i < numtiles; i++) {
		/* Get Tile Set Attributes */
		if (! tilecount) {
			attrib		= AttribTable[tileset];
			tilecount	= (short)TileSetSize (attrib);
			++tileset;
		}
		/** Separate Tile Type from actual tile data size. **/
		size = TileSize (tilesizes[i]);
		type = (short) TileUserType (tilesizes[i]);
		
		/** Read in Shape **/
		if (FPF_read (RawBuf, (size_t)size) != (size_t)size)
		{
			GEcatf ("\nReadInTiles:reading shape");
			goto CLOSE_PIX;
		}

		/**	Get pointer to copy of shape in current gr. mode format **/
			switch (type) {
			case 0: case 1: case 3: case 4:
				AttribTable[tileset-1] = (attrib & (~TSAT_FREE_CODE_MASK)) |
					TSAT_SHAPE_FREE;
#if __MSDOS__
				switch (GraphicsMode) {
				case MODE_EGA: case MODE_TANDY:
					cs.PatternMap = &cti->PatternInfo.PatmapTables[
						(patindex = (short)EGAPatIndex(attrib))];
					break;
				case MODE_CGA:
					cs.PatternMap = &cti->PatternInfo.PatmapTables[
						(short)CGAPatIndex(attrib)];
					break;
				}	
				/* Translate */
				cs.ExpandPatterns = (patindex != prevpatindex);
				prevpatindex = patindex;
				cs.pcc = (UBYTE)TilePCC (attrib);
#endif
				cs.ShapeData = (GShape *)RawBuf;
				if ( (newshapeptr = ConvertGShape (&cs)) == NULL)
				{
					goto CLOSE_PIX;
				}
				break;
			case 2: case 5: case 6:
				AttribTable[tileset-1] = (attrib & (~TSAT_FREE_CODE_MASK)) |
					TSAT_DIRECT_FREE;
				if ( (newshapeptr = GetCopy (RawBuf, (size_t)size)) == NULL)
				{
					SetGlobalErr ( ERR_OUT_OF_MEMORY);
					GEcatf ("IDAT3");
					goto CLOSE_PIX;
				}
				break;
			}

		/** Convert Size entry to Pointer **/
			((GShape **)tilesizes)[i] = newshapeptr;

		/** Set numtiles to the number of tiles sucessfully **/
		/** loaded so if failure occurs, FreeCTI won't free **/
		/** shapes that don't exist.						**/
		cti->NumTiles++;
	
		--tilecount;
	}

CLOSE_PIX:
	ConvertGShape (NULL); /* So it can shut down (free buffers etc.) */
	ClosePixlScan (&cs, MAX_COLORS);

	return result;

} /* ReadInTiles */

/*********************************************************************
 *
 * LowLoadCTI
 *
 * SYNOPSIS
 *		CTIType	*LowLoadCTI (
 *				char 	*filename,
 *				short		 fpfindex,
 *				CTIType	*globalcti,
 *				short		 isglobal
 *			)
 *		
 * USAGE
 *		CTIptr = LowLoadCTI ("ctipack.fpf", DISKMAN_F, NULL, FALSE);
 *		if (!CTIptr)
 *			printf ("LoadCTI Failed: %s\n", GlobalErrMsg);
 *
 * PURPOSE
 *		To load all the image data from a CTI (consolidated tile images) file
 *		and convert the images data upon loading into the appropriate graphic
 *		mode format.  To allocate and fill in a Recovery Table for 
 *		recovering tUME tile set user number and tile number from CTI index
 *		values.  To create image data lookup table to be indexed by CTI index
 *		values.  To allocate and fill in a Tile Sets Offsets Table for find-
 *		ing the start of a tile set in the lookup table by user number
 *
 * INPUTS
 *		filename:	Name of FilePack file containing CTI File to load.
 *		fpfindex:	FPF index to CTI file to load. 
 *		globalcti:	Pointer to already loaded Global Tiles CTI or NULL.
 *		isglobal:	Flag for whether this is the Global CTI being loaded.
 *	
 * EFFECTS
 *		When 'isglobal' is TRUE then tiles are loaded normally except that
 *		the TileShapes array is allocated big enough to hold the max number
 *		of Local tiles in addition to the Global tiles now being loaded.
 *		The max number of Local Tiles is taken to be the 'NumGlobalTiles'
 *		field.  The 'HasGlobals' field of the CTI structure is set to TRUE.
 *	
 *		When 'isglobal' is FALSE then 
 *		if 'globalcti' is not NULL then the space at the end of its 
 *			TilShapes pointer table will be used for the pointers to the 
 *			Local tiles to be loaded. The 'HasGlobals' field of
 *			the CTI structure is set to TRUE.
 *		Else This CTI does not use Global tiles so a TileShapes table just
 *			big enough for the Local tiles to be loaded will be allocated.
 *			The 'HasGlobals' field of the CTI structure is set to FALSE.
 *
 * RESULTS
 *		Returns pointer to CTI tracking structure if succesful, else returns
 *		NULL and sets GlobalErr and GlobalErrMsg.
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *	02/18/90 Sunday (JMA) - Made to work with Chunk output format and FPF
 *						storage.
 *	02/21/90 Wednesday (JMA) - Added support for global tile sharing.
 *
 * SEE ALSO
 *
*/
CTIType	*LowLoadCTI (
	char	*filename,
	short		 fpfindex,
	CTIType	*globalcti,
	short		 isglobal
)
{
	CTIType				*cti;
	ChunkType			 chunk;
	CTIHeaderType		 header;
	short	 			 GotIDAT = FALSE;
	UWORD				 numtiles;
	size_t				 size4;
#if __MSDOS__
	size_t				 sizetablesize, pattabarraysize = 0, totalpatdatsize;
	size_t				 spacetoalloc = 0;
	PatmapTableType		*patmaptableptr;
	UWORD				*tablesizes;
	UBYTE				*patdatptr, *patdatloadptr;
	UBYTE				*oddpatptr, *evenpatptr;
	UBYTE				 cgapat;
	UBYTE				 temp;
	short	 			 i, p;
	short				 numtables;
	UWORD				 numpats;
#endif
	ULONG				*tilesizes;
	size_t				 imagetablesize;
	ULONG				*AttribTable = NULL;

	/** Allocate CTI structure **/
		if ((cti = (CTIType *)MEM_calloc ((size_t)1, sizeof (CTIType))) == NULL) {
			SetGlobalErr ( ERR_OUT_OF_MEMORY);
			GEprintf2 ("LoadCTI 1:OOM allocating CTIType '%s'-%d\n", filename, fpfindex);
			return NULL;
		}

	/** Open File **/
		if (! FPF_open (filename, fpfindex)) {
			GEcatf ("\nLoadCTI 2:Opening file");
			goto FREE_EXIT;
		}


 	/** Read And Process Chunks Until Final Chunk (IDAT) is Read **/
	while (! GotIDAT) {
		/** Read Chunk Label and Size **/
		if ( (FPF_read (&chunk, sizeof (ChunkType))) !=
			sizeof (ChunkType)) 
		{
			GEcatf ("\nLoadCTI 3:Reading Chunk header");
			goto CLOSE_EXIT;
		}

		/** Identify and Process Chunk */
		if (chunk.ChunkLabel == CHUNK_HEAD) {		 /** H E A D **/	
			if ( (FPF_read (&header, sizeof (CTIHeaderType))) != 
				sizeof (CTIHeaderType))
			{
				GEcatf ("\nLoadCTI 3:reading HEAD chuck");
				goto CLOSE_EXIT;
			}
			if (!isglobal && globalcti) { /* check for errors */
				/* Global Tile Mismatch */
				if (globalcti->NumTiles != header.NumGlobalTiles) {
					SetGlobalErr ( ERR_GENERIC);
					GEprintf (GlobalErrStr[ERR_GENERIC]);
					GEcatf2 ("Global tile # mismatch in %s:%d", 
						filename, fpfindex);
					goto CLOSE_EXIT;
				}
				/* Max Local Tiles Exceeded */
				if (globalcti->NumGlobalTiles < header.NumTiles) {
					SetGlobalErr ( ERR_GENERIC);
					GEprintf (GlobalErrStr[ERR_GENERIC]);
					GEcatf4 ("Exeeded Max Local Tiles (%d > %d) in %s:%d", 
						header.NumTiles, globalcti->NumGlobalTiles,
						filename, fpfindex);
					goto CLOSE_EXIT;
				}
			}
			cti->HasGlobals = (isglobal || globalcti);
			numtiles = header.NumTiles;
			cti->NumGlobalTiles = header.NumGlobalTiles;
			cti->NumGlobalTileSets = header.NumGlobalTileSets;
		} else if (chunk.ChunkLabel == CHUNK_RTAB) {	/** R T A B **/
			
			/* Allocate Recover Table */
			if ( (cti->RecoverTable = (UWORD *)MEM_malloc (
				(size_t)header.RecoverTableSize)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf (GlobalErrStr[ERR_OUT_OF_MEMORY]);
				GEcatf1 ("'%s' RTAB", filename);
				goto CLOSE_EXIT;
			}
			/* Read in Recover Table */
			if (FPF_read (cti->RecoverTable, 
				(size_t)header.RecoverTableSize) != header.RecoverTableSize)
			{
				GEcatf ("\nLoadCTI:Reading RTAB(2)");
				goto CLOSE_EXIT;
			}
		
		} else if (chunk.ChunkLabel == CHUNK_OTAB) { /** O T A B **/

			/* Allocate Tile Set Offsets Table */
			if ((cti->TileSetOffsets = (WORD *)MEM_malloc (
				(size_t)header.OffsetTableSize)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadCTI:OOM Offset table in '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
	 		}

			/* Read in Tile Set Offsets Table */
			if (FPF_read (cti->TileSetOffsets, 
				(size_t)header.OffsetTableSize) != header.OffsetTableSize)
			{
				GEcatf ("\nLoadCTI:reading OTAB(2)");
				goto CLOSE_EXIT;
			}
		}
#if __MSDOS__
		 else if ((chunk.ChunkLabel == CHUNK_EPRM &&
						(GraphicsMode == MODE_EGA || 
					 		GraphicsMode == MODE_TANDY)) ||
					(chunk.ChunkLabel == CHUNK_CPRM &&
						GraphicsMode == MODE_CGA))	{	/** E P R M **/
														/** C P R M **/
			numtables 
			= cti->PatternInfo.NumPatmapTables 
			= (chunk.ChunkLabel == CHUNK_EPRM)? header.NumEGAPatmapTables :
				header.NumCGAPatmapTables;
			pattabarraysize = numtables * sizeof (PatmapTableType);

			/* Allocate Buffer for Patmap Table Structures */
			if ((patmaptableptr = (PatmapTableType *)CreateTempBuffer 
				(pattabarraysize)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadCTI:OOM PatmapTableType in '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
			cti->PatternInfo.PatmapTables = patmaptableptr;

			/** Read Patmap Table Sizes into Low End of Buffer **/
			sizetablesize		= numtables * sizeof (UWORD);
			tablesizes			= (UWORD *) patmaptableptr;
			tablesizes			= (UWORD *) ((UBYTE *) tablesizes + 
								   (pattabarraysize - sizetablesize));

			if ( (FPF_read (tablesizes, sizetablesize) != sizetablesize)) 
			{
				GEcatf ("\nLoadCTI:reading ?PRM chunk");
				goto CLOSE_EXIT;
			}

			/** Get Total size of EGA Patmap Tables **/
			spacetoalloc = 
 				totalpatdatsize = 
					(size_t)chunk.ChunkSize - sizetablesize;

			/* Double for CGA so can unpack each pat byte into two bytes */
			if (chunk.ChunkLabel == CHUNK_CPRM ) {
				spacetoalloc *= 2;
			}

			/** Allocate Buffer2 for All EGA Patmap Table Data **/
			if ((patdatptr = (UBYTE *)CreateTempBuffer (spacetoalloc)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadCTI:OOM patmap table data '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
			cti->PatternInfo.DataBuf = patdatloadptr = patdatptr;

			/** For CGA, read patterns into lower half of buffer **/
			if (chunk.ChunkLabel == CHUNK_CPRM ) {
				patdatloadptr += totalpatdatsize;
			} 

			/** Read All Patmap Tables into Buffer **/
			if ((FPF_read (patdatloadptr, totalpatdatsize)) != 
				totalpatdatsize) 
			{
				MEM_free (patdatptr);
				GEcatf ("\nLoadCTI:reading ?PRM(4) chunk");
				goto CLOSE_EXIT;
			}


			/** Set Pointers in PatmapTable Structs to Data in Buffer2 **/
			for (i = numtables; i; i--) {
				numpats = *tablesizes++;

				/* Unpack and move CGA pattern arrays */
				if (chunk.ChunkLabel == CHUNK_CPRM) {

					/* Get src and dest ptrs for unpacking CGA patterns */
					oddpatptr = patdatptr;
					evenpatptr = patdatptr + numpats;

					for (p = numpats; p ; p--) {
						cgapat = *patdatloadptr++; /* Get packed pat byte */

						/* Copy odd line pat from low nybble */
						temp = cgapat & 0x0F;
						*oddpatptr++ = temp | (temp << 4);

						/* Copy even line pat from high nybble */
						temp = cgapat & 0xF0;
						*evenpatptr++ = temp | (temp >> 4);
					}
				} else { /* EGA pats are a word */
					numpats /= 2; /* size of table / size of pattern */
				}
				patmaptableptr->NumPatterns = numpats;
				patmaptableptr->Odd = patdatptr;
				patdatptr += numpats;
				patmaptableptr->Even = patdatptr;
				patdatptr += numpats;
				++patmaptableptr;
			}

		}
#endif 
		else if (chunk.ChunkLabel == CHUNK_TSAT)	{ /** T S A T **/
			size4 = header.NumTileSets * sizeof (ULONG);
			if (!isglobal && globalcti) { /* Use Global CTI Table Space */
				AttribTable = ((ULONG *)globalcti->TileSetAttributes) + 
					globalcti->NumGlobalTileSets;
			} else {
				if (isglobal) { /* Add in Space for Max Local Tile Sets */
					size4 += header.NumGlobalTileSets * sizeof (ULONG);
				}
				/** Allocate Tile Set Attributes Table **/
				if ((AttribTable = (ULONG *)MEM_malloc (size4)) == NULL)
				{
					SetGlobalErr ( ERR_OUT_OF_MEMORY);
					GEprintf2 ("LoadCTI:OOM Attribute table '%s'-%d", 
						filename, fpfindex);
					goto CLOSE_EXIT;

				}
				cti->TileSetAttributes = AttribTable;
			}

			/** Read In Tile Set Attributes **/
			if ((FPF_read (AttribTable, (size_t)chunk.ChunkSize)) 
				!= (size_t)chunk.ChunkSize) 
			{
				GEcatf ("\nLoadCTI:reading tile attributes TSAT(2)");
				goto CLOSE_EXIT;
			}
		} else if (chunk.ChunkLabel == CHUNK_ISIZ)	{ /** I S I Z **/
			imagetablesize = numtiles * sizeof (ULONG);
			/* Add one to size allocated so we can index from 1..NumTiles */
			size4 = imagetablesize + sizeof (ULONG);
			if (!isglobal && globalcti) { /* Use Global CTI Table Space */
				tilesizes = ((ULONG *)globalcti->TileShapes) + 
					globalcti->NumTiles;
				cti->TileShapes = globalcti->TileShapes;

			} else {
				if (isglobal) { /* Add in Space for Max Local Tiles */
					size4 += header.NumGlobalTiles * sizeof (ULONG);
				}
				/** Allocate Tile Image Sizes Table **/
				if ((tilesizes = (ULONG *)MEM_malloc (size4)) == NULL)
				{
					SetGlobalErr ( ERR_OUT_OF_MEMORY);
					GEprintf2 ("LoadCTI:OOM Size table '%s'-%d", filename, fpfindex);
					goto CLOSE_EXIT;

				}
				cti->TileShapes = (GShape **)tilesizes;
			}

			/* Add one to tilesizes ptr so we can index from 1..NumTiles */
			++tilesizes;

			/** Read in Image Sizes Table **/
			if (FPF_read (tilesizes, imagetablesize) != imagetablesize)
			{
				GEcatf ("\nLoadCTI:reading image sizes");
				goto CLOSE_EXIT;
			}
		} else if (chunk.ChunkLabel == CHUNK_IDAT)	{ /** I D A T **/
			GotIDAT = TRUE;
			/** Get Shape Processing Buffers **/
			if ( (RawBuf = (UBYTE *)GetGShapeLoadBuf ()) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadCTI:OOM getting rawbuf '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}

			if (! ReadInTiles (numtiles, cti, AttribTable, tilesizes)) {
				goto CLOSE_EXIT;
			}
		} else { /* Unknown or UnNeeded Chunk, Skip It */
			if (! FPF_skip (chunk.ChunkSize)) {
				SetGlobalErr ( ERR_READING_FILE);
				GEprintf2 ("LoadCTI: unknown chuck type/size '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
		}
	}

#if __MSDOS__
	if (spacetoalloc)		DestroyTempBuffer (patdatptr, spacetoalloc);
	if (pattabarraysize)	DestroyTempBuffer (patmaptableptr, pattabarraysize);
#endif
	FPF_close ();
	return cti;

CLOSE_EXIT:
	FPF_close ();
FREE_EXIT:
#if __MSDOS__
	if (spacetoalloc)		DestroyTempBuffer (patdatptr, spacetoalloc);
	if (pattabarraysize)	DestroyTempBuffer (patmaptableptr, pattabarraysize);
#endif
	FreeCTI (cti);
	return NULL;
} /* LowLoadCTI */

/*********************************************************************
 *
 * LowFreeCTI
 *
 * SYNOPSIS
 *		void	LowFreeCTI (
 *			CTIType	*cti,
 *			short		 freeglobal
 *		)	
 *
 * PURPOSE
 *		To free the memory allocted by LoadCTI for the tile images data and
 *		associated tables (recover and offset).
 *		Note: If the 'HasGlobals' field is FALSE then the CTI could not
 *		possibly be the Global Tiles CTI.  If it is TRUE then it could be the
 *		Global Tiles CTI or a CTI which uses the Global Tiles CTI shape 
 *		pointer table.
 *
 * INPUT
 *		cti:	Tracking structure created by LoadCTI().
 *
 * EFFECTS
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void	LowFreeCTI (
	CTIType	*cti,
	short		 freeglobal
)	
{
	short	i, start = 0, end = cti->NumTiles;
	short	tileset, tilecount, tilefreecode;
	ULONG	attrib;

	if (!cti)
		return;

	if (cti->HasGlobals) {
		if (!freeglobal) { /* Used the Global CTI shape pointer table */
			start = cti->NumGlobalTiles;
			end = cti->NumTiles + start;
		} /* else This is the Global Tiles CTI */
	} /* This is a stand alone CTI, No global tiles */

	/* Free Tile Shapes and Masks */
	tileset		= 0;
	tilecount	= 0;
	for (i = start; i < end; i++) {

		/* Get Tile Set Attribute */
		if (! tilecount) {
			attrib			= cti->TileSetAttributes[tileset];
			tilecount		= (short)TileSetSize (attrib);
			tilefreecode	= (short)TileFreeCode (attrib);
			++tileset;
		}

		switch (tilefreecode) {
		case TSAT_DIRECT_FREE:
			if (cti->TileShapes[i + 1]) {
				MEM_free (cti->TileShapes[i + 1]);
			}
			break;
		case TSAT_SHAPE_FREE:
			if (cti->TileShapes[i + 1]) {
				FreeGShape (cti->TileShapes[i + 1]);
			}
			break;
		}
		--tilecount;
	}

	/* Free Tile Shape and Mask Pointer Array */
	if (((cti->HasGlobals) ? freeglobal : TRUE) && cti->TileShapes)
		MEM_free (cti->TileShapes);

	/* Free Tile Set Attributes Array */
	if (cti->TileSetAttributes)
		MEM_free (cti->TileSetAttributes);

	/** Free PatternInfo **/
#if __MSDOS__
	switch (GraphicsMode) {
	case MODE_EGA: case MODE_TANDY: case MODE_CGA:
		if (cti->PatternInfo.PatmapTables) {
			if (cti->PatternInfo.DataBuf) {
				MEM_free (cti->PatternInfo.DataBuf);
			}
			MEM_free (cti->PatternInfo.PatmapTables);
		}
		break;
	}
#endif

	/* Free Recover Table */
	if (cti->RecoverTable) {
		MEM_free (cti->RecoverTable);
	}

	/* Free Tile Set Offsets Table */
	if (cti->TileSetOffsets) {
			MEM_free (cti->TileSetOffsets);
	}

	MEM_free (cti);
} /* LowFreeCTI */

/*********************************************************************
 *
 * LoadtUMEPackRoom
 *
 * SYNOPSIS
 *		tUMEPackRoomType	*LoadtUMEPackRoom	(
 *			char				*filename,
 *			short					 fpfindex
 *			CTIType				*cti,
 *			tUMEPackRoomType	*tproomptr,
 *			short				 numcopies;
 *		)
 *
 * USAGE
 *		tprroomptr = LoadtUMEPackRoom ("room00.001", tproomptr);
 *		if (!tproomptr)
 *			printf ("Error loading Game Map\n");
 *
 * PURPOSE
 *		To load in a game room from disk.  The first time this routine is 
 *		called you should pass a NULL pointer for 'tproomptr' and a 
 *		tUMEPackRoomType struct will be allocated and filled in.  Thereafter
 *		call the routine with a pointer to that structure and it will be
 *		re-used along with the memory allocated for the room layers.
 *		However, do not re-use a structure to load a room that is larger than
 *		the room of the previous load (or memory will get trashed).
 *	
 * INPUTS
 *		filename: 		Name of FilePack File containing tumepack created 
 *						GameRoom file.
 *		fpfindex:		Index to GameRoom file in FilePack File.
 *		cti:			CTI used with room to be loaded (YOU MUST PASS THIS!)
 *		tproomptr:		pointer to structure with fields for relavent 
 *						parameters of a room:
 *							WORD			 RoomWidth;
 *							WORD			 RoomHeight;
 *							WORD			 NumLayers;
 *							WORD			 TileWidth;
 *							WORD			 TileHeight;
 *							UBYTE			 NumRoomCopies;
 *							PaletteType		 Palette;
 *							CTIType			*CTI;
 *							RoomType		*Rooms;
 *							char			*Name;
 *		numcopies:		Number of copies of  the room to make.
 *		
 * RESULTS
 *		Returns pointer to tUMEPackRoomType struct if succesful. Else returns
 *		NULL.
 *
 * BUGS
 *
 *
 * HISTORY
 *	11/30/89 Thursday - Converted from JMA's LoadGameMap to include
 *	                    support for ALLOCFLAG. (RGM)
 *
 *	01/16/90 Tuesday  - Changed routine so that it allocates each
 *						item that is not yet allocated. (GAT)
 *	02/20/90 Tuesday (JMA) - Made to work with Chunk output format and FPF
 *						storage.  Turned static room array into dynamically
 *						allocated array (so you can get as many copies as
 *						you want now).
 *	03/08/90 Thursday (JMA) - Added Name filed to room structure. Made it a
 *					staticly allocated array of size ROOMNAME_MAXLEN.  Only
 *					up to ROOMNAME_MAXLEN chars will be read.
 *
 * SEE ALSO
 *
*/
tUMEPackRoomType	*LoadtUMEPackRoom (
	char				*filename,
	short					 fpfindex,
	CTIType				*cti,
	tUMEPackRoomType	*tproomptr,
	short				 numcopies
)
{
	short				i, r;
	ChunkType			chunk;
	TPRoomHeaderType	header;
	short				GotLDAT = FALSE;
	size_t				layerbytesize;
	size_t				layertablesize;
	short				allocflag = TRUE;
	size_t				namesize;
	UBYTE				foundMPAL = FALSE;

	/** Allocate tUMEPackRoomType struct if necessary **/
		if (!tproomptr) {
			if ((tproomptr = (tUMEPackRoomType *)MEM_calloc ((size_t)1, sizeof (
					tUMEPackRoomType))) == NULL) {
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf (GlobalErrStr[ERR_OUT_OF_MEMORY]);
				GEcatf1 ("LoadtPRoom 1: '%s'", filename);
				return NULL;
				
			}
		} else {
			allocflag = FALSE;
		}

	/** Open File **/
		if (! FPF_open (filename, fpfindex)) {
			GEcatf ("\nLoadtPRoon 2:Opening file");
			goto FREE_EXIT;
		}


 	/** Read And Process Chunks Until Final Chunk (LDAT) is Read **/
	while (! GotLDAT) {
		/** Read Chunk Label and Size **/
		if ( (FPF_read (&chunk, sizeof (ChunkType))) !=
			sizeof (ChunkType)) 
		{
			GEcatf ("\nLoadtPRoom 3:reading CHUCK header");
			goto CLOSE_EXIT;
		}

		/** Identify and Process Chunk */
		if (chunk.ChunkLabel == CHUNK_HEAD) {		 /** H E A D **/	
			if ( (FPF_read (&header, sizeof (TPRoomHeaderType))) != 
				sizeof (TPRoomHeaderType))
			{
				GEcatf ("\nLoadtPRoom 4:reading HEAD chunk");
				goto CLOSE_EXIT;
			}
			/* Copy RoomWidth,RoomHeight,NumLayers,TileWidth,TileHeight */
			*((TPDimHeaderType *)tproomptr) = *((TPDimHeaderType *)&header);

#if __MSDOS__
			if (GraphicsMode == MODE_CGA) {
				tproomptr->Palette.NumColors = NUM_CGA_COLORS;
				tproomptr->Palette.Data.CGA = header.CGAPaletteCode;
			}
#endif

		} else if (chunk.ChunkLabel == CHUNK_NAME)	{ /** N A M E **/
			namesize = min ((size_t)chunk.ChunkSize, ROOMNAME_MAXLEN);
			/* Read Name */
			if ( (FPF_read (tproomptr->Name, 
				(size_t)namesize)) != namesize) 	
			{
				GEcatf ("\nLoadtPRoom 5:reading NAME chunk");
				goto CLOSE_EXIT;
			}
			if (namesize < chunk.ChunkSize) { /* skip the rest */
				if (! FPF_skip (chunk.ChunkSize - namesize)) 
				{
					GEcatf ("\nLoadtPRoom 6:skip rest of NAME chunk");
					goto CLOSE_EXIT;
				}
			}
		} else if (chunk.ChunkLabel == CHUNK_MPAL) { /** M P A L **/
			foundMPAL = TRUE;
#if __MSDOS__
			if (GraphicsMode == MODE_MCGA)
#endif
			{
				/* Allocate RGB Palette */
				if (!tproomptr->Palette.Data.RGB) {
					if ((tproomptr->Palette.Data.RGB = (ColorType *)MEM_malloc (
						(size_t)chunk.ChunkSize)) == NULL)
					{
						SetGlobalErr ( ERR_OUT_OF_MEMORY);
						GEprintf2 ("LoadtPRoom 7:OOM mpalette MPAL '%s'-%d", filename, fpfindex);
						goto CLOSE_EXIT;
					}
				}
				/* Read Palette */
				if ( (FPF_read (tproomptr->Palette.Data.RGB, 
					(size_t)chunk.ChunkSize)) !=	chunk.ChunkSize) 	
				{
					MEM_free (tproomptr->Palette.Data.RGB);
					SetGlobalErr ( ERR_READING_FILE);
					GEprintf (GlobalErrStr[ERR_READING_FILE]);
					GEcatf1 ("'%s' tMPAL2\n", filename);
					goto CLOSE_EXIT;
				}
				tproomptr->Palette.NumColors = header.NumMCGAColors;
			}
#if __MSDOS__
			 else {
				if (! FPF_skip (chunk.ChunkSize)) {
					SetGlobalErr ( ERR_READING_FILE);
					GEprintf2 ("LoadtPRoom:Error skipping MPAL/size:'%s'-%d", filename, fpfindex);
					goto CLOSE_EXIT;
				}
			}
#endif
		}
#if __MSDOS__
		 else if (chunk.ChunkLabel == CHUNK_EPAL &&
					(GraphicsMode == MODE_EGA ||
					 GraphicsMode == MODE_TANDY))	{ /** E P A L **/
			/* Allocate EGA Palette */
			if (!tproomptr->Palette.Data.EGA) {
				if ((tproomptr->Palette.Data.EGA = MEM_malloc ((size_t)chunk.ChunkSize)) == NULL)
				{
					SetGlobalErr ( ERR_OUT_OF_MEMORY);
					GEprintf (GlobalErrStr[ERR_OUT_OF_MEMORY]);
					GEcatf1 ("'%s' tEPAL\n", filename);
					goto CLOSE_EXIT;
				}
			}
			/* Read Palette */
			if ( (FPF_read (tproomptr->Palette.Data.EGA, 
				(size_t)chunk.ChunkSize)) != chunk.ChunkSize) 
			{
				MEM_free (tproomptr->Palette.Data.EGA);
				SetGlobalErr ( ERR_READING_FILE);
				GEprintf (GlobalErrStr[ERR_READING_FILE]);
				GEcatf1 ("'%s' tEPAL2\n", filename);
				goto CLOSE_EXIT;
			}
			tproomptr->Palette.NumColors = header.NumEGAColors;
		}
#endif 
		else if (chunk.ChunkLabel == CHUNK_LDAT)	{ /** L D A T **/
			GotLDAT = TRUE;
			/* Allocate Room Pointers */
			if (allocflag) {
				if ( (tproomptr->Rooms = (RoomType *)MEM_malloc (
					(numcopies+1) * sizeof(RoomType))) == NULL)
				{
					SetGlobalErr ( ERR_OUT_OF_MEMORY);
					GEprintf (GlobalErrStr[ERR_OUT_OF_MEMORY]);
					GEcatf1 ("'%s' LDAT\n", filename);
					goto CLOSE_EXIT;
				} 
			}
				
			layertablesize = (size_t)(header.NumLayers) * sizeof (WORD *);
			layerbytesize = (size_t)(header.RoomWidth) * 
					(size_t)(header.RoomHeight) * sizeof (WORD);

			if (allocflag) { /* Allocate Layer Pointer Table */
				if ((tproomptr->Rooms[0] = (RoomType)MEM_calloc ((size_t)1,
							layertablesize)) == NULL)
				{
					SetGlobalErr ( ERR_OUT_OF_MEMORY);
					GEprintf (GlobalErrStr[ERR_OUT_OF_MEMORY]);
					GEcatf1 ("'%s' LDAT2\n", filename);
					goto CLOSE_EXIT;
				} 
			} 

			/** Read in Layers (and allocate them if allocflag is TRUE) **/
			for (i = 0; i < header.NumLayers; i++) {
				if (allocflag) { /* Allocate storage for a layer */
					if ((tproomptr->Rooms[0][i] = (LayerType)MEM_malloc (
								layerbytesize)) == NULL)
					{
						SetGlobalErr ( ERR_OUT_OF_MEMORY);
						GEprintf (GlobalErrStr[ERR_OUT_OF_MEMORY]);
						GEcatf1 ("'%s' LDAT3\n", filename);
						goto CLOSE_EXIT;
					}
				} 

				/* Read layer data into storage */
				if (FPF_read (tproomptr->Rooms[0][i], 
					layerbytesize) != layerbytesize)
				{
					SetGlobalErr ( ERR_READING_FILE);
					GEprintf (GlobalErrStr[ERR_READING_FILE]);
					GEcatf1 ("'%s'LDAT4\n", filename);
					goto CLOSE_EXIT;
				}
			}

			/** Make As Many Copies As Asked For */
			tproomptr->NumRoomCopies = numcopies;
			for (r = 1; r <= numcopies; r++) {
				if (allocflag) { /* Allocate Layer Pointer Table */
					if ((tproomptr->Rooms[r] = (RoomType)MEM_calloc ((size_t)1,
									layertablesize)) == NULL) {
						SetGlobalErr ( ERR_OUT_OF_MEMORY);
						GEprintf (GlobalErrStr[ERR_OUT_OF_MEMORY]);
						GEcatf1 ("'%s' LDAT5\n", filename);
						goto CLOSE_EXIT;
					}
				}
				for (i = 0; i < header.NumLayers; i++) {
					if (allocflag) { /* Allocate storage for a layer */
						if ((tproomptr->Rooms[r][i] = (LayerType)MEM_malloc (
									layerbytesize)) == NULL) {
							SetGlobalErr ( ERR_OUT_OF_MEMORY);
							GEprintf (GlobalErrStr[ERR_OUT_OF_MEMORY]);
							GEcatf1 ("'%s' LDAT6\n", filename);
							goto CLOSE_EXIT;
						}
					}
					memcpy (tproomptr->Rooms[r][i], tproomptr->Rooms[0][i], 
							layerbytesize);
				}
			}
		} else { /* Unknown or UnNeeded Chunk, Skip It */
			if (! FPF_skip (chunk.ChunkSize)) 
			{
				SetGlobalErr ( ERR_READING_FILE);
				GEprintf (GlobalErrStr[ERR_READING_FILE]);
				GEcatf1 ("LoadtPRoom 4:'%s'\n", filename);
				goto CLOSE_EXIT;
			}
		}
	}

	FPF_close ();

#if __MSDOS__
	if (foundMPAL && 
		(GraphicsMode == MODE_EGA || GraphicsMode == MODE_TANDY) &&
		!tproomptr->Palette.Data.EGA) {

		short	temp;

		if ((tproomptr->Palette.Data.EGA = MEM_malloc (16)) == NULL) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEprintf2 ("LoadBPI:OOM allocating EGA fake palette:'%s'-%d", filename, fpfindex);
			goto CLOSE_EXIT;
		}
		for (temp = 0; temp < 16; temp++) {
			tproomptr->Palette.Data.EGA[temp] = temp;
		}
		tproomptr->Palette.NumColors = 16;
	}
#endif

	if (cti) {					/* only reset the cti ptr if	*/
		tproomptr->CTI = cti;	/* one was passed in			*/
	}

	return tproomptr;

CLOSE_EXIT:
	FPF_close ();
FREE_EXIT:
	FreetUMEPackRoom (tproomptr);
	return NULL;

} /* LoadtUMEPackRoom */

/*********************************************************************
 *
 * FreetUMEPackRoom
 *
 * SYNOPSIS
 *		void	FreetUMEPackRoom (
 *			tUMEPackRoomType *tproomptr
 *		) 
 *
 * PURPOSE
 *		To free the memory allocated by LoadtUMEPackRoom().
 *
 * INPUT
 *
 * EFFECTS	
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		LoadtUMEPackRoom()
*/
void	FreetUMEPackRoom (
	tUMEPackRoomType	*tproomptr
) 
{
	short i, r;

	if (!tproomptr)
		return;

	for (r = 0; r < tproomptr->NumRoomCopies; r++) {
		/* Free all layers of Room */
		for (i = 0; i < tproomptr->NumLayers; i++) {
			if (tproomptr->Rooms[r][i])
				MEM_free (tproomptr->Rooms[r][i]);
		}
		/* Free Table of Pointers to Layers */
		if (tproomptr->Rooms[r])
			MEM_free (tproomptr->Rooms[r]);
	}
	
	/* Free Room pointer table */
	if (tproomptr->Rooms) {
		MEM_free (tproomptr->Rooms);
	}

	/** Free Palette **/
	if (tproomptr->Palette.NumColors) {
#if __MSDOS__
		switch (GraphicsMode) {
		case MODE_MCGA:
			MEM_free (tproomptr->Palette.Data.RGB);
			break;
		case MODE_EGA: case MODE_TANDY:
			MEM_free (tproomptr->Palette.Data.EGA);
			break;
		}
#endif
#if __AMIGAOS__
#endif
	}

	MEM_free (tproomptr);
} /* FreetUMEPackRoom */


/*********************************************************************
 *
 * LoadConversionRoom
 *
 * SYNOPSIS
 *		ConversionRoomType	*LoadConversionRoom (
 *			char			*filename
 *		)
 *
 * PURPOSE
 *		To load a 'conversion room' from a file into a conversion
 *		table structure.
 *
 * INPUT
 *		char *filename		: pointer to filename string.
 *
 * EFFECTS
 *		Allocates a new conversion room and fills in the values from
 *		the conversion room file. NO CLEANUP ON FAILURE.
 *
 * RETURN VALUE
 *		A pointer to the newly created conversion room, or NULL if
 *		failed.
 *
 * HISTORY
 *	12/18/89 Monday   - Created. (RGM)
 *
 *	12/21/89 Thursday - Converted to true non-homogeneous conversion
 *	                    room support. (RGM)
 *
 *	01/09/90 Tuesday -	Changed routine to use open() instead of fopen()
 *						for speed. (GAT)
 *	
 *	02/20/90 Tuesday (JMA) - Made to work with FPF storage.  
 *
 * SEE ALSO
 *		FreeConversionRoom ().
*/

ConversionRoomType *LoadConversionRoom (
	char	*filename,
	short		 fpfindex
)
{
	ConversionRoomType	*conroom;
	ConTablePtrType		contable;
	short				count;
	short				conversionsize;
	short				tablecount;
	short				items_read;
	short				layers;
	short				entries;
	short				total_entries;
	UWORD				*existance_table;
	UBYTE				*highdata;
	UBYTE				*lowdata;
	UBYTE				*conversionchunk;
	short				layerindex;

#if REQUIRE
	/* REQUIRE valid arguments */
	if (filename == NULL) {
		SetGlobalErr ( ERR_REQUIREMENTS);
		GEprintf (GlobalErrStr [GlobalErr]);
/**/	goto ABORT;
	}
#endif /* REQUIRE */

	/* try to get a file handle */
	if (! FPF_open (filename, fpfindex)) {
		SetGlobalErr ( ERR_FILE_NOT_FOUND);
		GEprintf (GlobalErrStr [GlobalErr]);
		GEcatf1 (" with file '%s',", filename); 
/**/	goto ABORT;
	}

	/* attempt a conversion room structure allocation */
	if ( (conroom = MEM_malloc (sizeof (ConversionRoomType))) == NULL) {
		SetGlobalErr ( ERR_OUT_OF_MEMORY);
		GEprintf (GlobalErrStr [GlobalErr]);
/**/	goto ABORT;
	}

	/* fill in the conversion room's header information */
	items_read = (short) FPF_read (conroom, sizeof(UWORD) * 2);
	if (items_read != sizeof(UWORD) * 2) {
		SetGlobalErr ( ERR_READING_FILE);
		GEprintf (GlobalErrStr [GlobalErr]);
		GEcatf1 (" with file '%s',", filename); 
/**/	goto ABORT;
	}

	/* make sure that we have a valid header */
	if (! conroom->Entries) {
		SetGlobalErr ( ERR_INVALID_DATA);
		GEprintf (GlobalErrStr [GlobalErr]);
		GEcatf1 (" with file '%s',", filename); 
/**/	goto ABORT;
	}

	/* try to allocate a conversion table pointer array */
	if ( (conroom->ConTable =
				MEM_malloc (conroom->Entries * sizeof (ConTablePtrType))) == NULL) {
		SetGlobalErr ( ERR_OUT_OF_MEMORY);
		GEprintf (GlobalErrStr [GlobalErr]);
/**/	goto ABORT;
	}

	tablecount = conroom->Entries;

	/* ughh, I don't exactly know how I'm going to do this next part, */
	/* so I'll adlib. . . */

	/* first load in the existance table temporarily into */
	/* the conversion table pointer array.  We're going to load it */
	/* at the halfway point of the table because we want to convert the */
	/* sizes to pointers IN PLACE. . .this is because */
	/* the size entries are UWORDS (2 bytes), while the contable */
	/* pointers are forced to FAR pointer (4 bytes). */

	existance_table = (UWORD *) conroom->ConTable;
	existance_table += tablecount;	/* gets us to the second half */
	
	items_read = (short) FPF_read (existance_table, sizeof(UWORD) *	tablecount);
	if (items_read != sizeof(UWORD) * tablecount) {
		SetGlobalErr ( ERR_READING_FILE);
		GEprintf (GlobalErrStr [GlobalErr]);
		GEcatf1 (" with file '%s',", filename); 
/**/	goto ABORT;
	}

	/* now we find out the total size of all the conversion data left in */
	/* the file and load it in */
	conversionsize = 0;
	for (count = 0; count < tablecount; count ++) {
		conversionsize += existance_table[count];
	}
	if ( (conversionchunk = (void *) MEM_malloc ((size_t)conversionsize)) == NULL) {
		SetGlobalErr ( ERR_OUT_OF_MEMORY);
		GEprintf (GlobalErrStr [GlobalErr]);
/**/	goto ABORT;
	}
	/* kludge to track the entire conversion chunk for later disposal */
	conroom->MemBuff = conversionchunk;

	items_read = (short) FPF_read (conversionchunk, sizeof(UBYTE) * conversionsize);
	if (items_read != sizeof (UBYTE) * conversionsize) {
		SetGlobalErr ( ERR_READING_FILE);
		GEprintf (GlobalErrStr [GlobalErr]);
		GEcatf1 (" with file '%s',", filename); 
/**/	goto ABORT;
	}

	/* now we allocate the conversion table structures we'll need. */
	/* Also, we'll fill in the data for the conversion data structure */
	/* from the data in the conversionchunk */
	for (count = 0; count < tablecount; count ++) {
		if (existance_table[count] == 0) {
			contable = (ConTablePtrType) NULL;
		}
		else {
			/* attempt a conversion room structure allocation */
			if ( (contable =
					(ConTablePtrType) MEM_malloc (sizeof (ConversionTableType)))==NULL){
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf (GlobalErrStr [GlobalErr]);
/**/			goto ABORT;
			}
			/* initialize the structure with data in the current conversion */
			/* chunk segment.  The actual header data in the conversion */
			/* chunk will not be used again */
			contable->Layers =
				((ConversionTableType *) conversionchunk)->Layers;
			contable->Entries =
				((ConversionTableType *) conversionchunk)->Entries;
			contable->Adjustment =
				((ConversionTableType *) conversionchunk)->Adjustment;

			/* make sure that we have a valid header */
			if ( (! contable->Entries) || (! contable->Layers) ) {
				SetGlobalErr ( ERR_INVALID_DATA);
				GEprintf (GlobalErrStr [GlobalErr]);
				GEcatf1 (" with file '%s',", filename); 
/**/			goto ABORT;
			}
			layers			= contable->Layers;
			entries			= contable->Entries;
			total_entries	= layers * entries;

			/* try to allocate the High & Low arrays */
			if ( (contable->High = MEM_malloc (layers * sizeof (UBYTE *))) == NULL) {
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf (GlobalErrStr [GlobalErr]);
/**/			goto ABORT;
			}
			if ( (contable->Low = MEM_malloc (layers * sizeof (UBYTE *))) == NULL) {
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf (GlobalErrStr [GlobalErr]);
/**/			goto ABORT;
			}

			/* assign the high and low bytes data pointer.  The high */
			/* pointer is the start of the current chunk segment */
			/* + 3 * sizeof (UWORD) to bypass the header information */
			/* (Layers, Entries, and Adjustment) */
			highdata = conversionchunk + (3 * sizeof (UWORD));
			lowdata = highdata + total_entries;

			/* set up the 'layer' pointers */
			for (layerindex = 0; layerindex < layers; layerindex += 1) {
				contable->High [layerindex] = highdata;
				contable->Low [layerindex] = lowdata;

				highdata += entries;
				lowdata  += entries;
			}

		} /* end else */

		conroom->ConTable[count] = contable;

		/* get us ready for the next conversion table */
		conversionchunk += existance_table[count];

	} /* end for */

	FPF_close ();
	return (conroom);
/*-------------------------------------------------------------------*/
ABORT:
	GEcatf ("\n   in function 'LoadConversionRoom' (RGM).\n");
	return (NULL);
} /* *LoadConversionTable */


/*********************************************************************
 *
 * FreeConversionRoom
 *
 * SYNOPSIS
 *		void FreeConversionRoom (
 *			ConversionRoomType	*conroom;
 *		)
 *
 * PURPOSE
 *		To free up a conversion room structure.
 *
 * INPUT
 *		A pointer to a conversion room.
 *
 * EFFECTS
 *		De-allocates the conversion table structures and the LOW and HIGH
 *		arrays.
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *	12/19/89 Tuesday - Created. (RGM)
 *
 *	12/21/89 Thursday - Converted from FreeConversionTable to include
 *                      support for true non-homogeneous conversion
 *                      rooms. (RGM)
 * SEE ALSO
 *		LoadConversionRoom ().
*/
void FreeConversionRoom (
	ConversionRoomType	*conroom
)
{
	short index;

#if REQUIRE
	/* REQUIRE that we have valid arguments */
	if (conroom == NULL) {
		SetGlobalErr ( ERR_REQUIREMENTS);
		goto ABORT;
	}
#endif /* REQUIRE */

	SAFEFREE (conroom->MemBuff);
	for (index = conroom->Entries - 1; index >= 0; index --) {
		if (conroom->ConTable[index]) {
			SAFEFREE (conroom->ConTable[index]->High);
			SAFEFREE (conroom->ConTable[index]->Low);
			SAFEFREE (conroom->ConTable[index]);
		}
	}
	SAFEFREE (conroom->ConTable);
	SAFEFREE (conroom);
/*-------------------------------------------------------------------*/
ABORT:
	return;
} /* FreeConversionTable */

/*********************************************************************
 *
 * DrawPartOfRoom
 *
 * SYNOPSIS
 *		void DrawPartOfRoom (
 *			tUMEPackRoomType	*tproom;
 *			short				 roomnumber;
 *			short				 xoff;		* X Pixel offset of drawing		*
 *			short				 yoff;		* Y Pixel offset of drawing		*
 *			short				 txoff;		* X tile offset to first tile	*
 *			short				 tyoff;		* Y tile offset to first tile	*
 *			short				 twidth;	* num of tiles to draw across	*
 *			short				 theight;	* num of tiles to draw down		*
 *		)
 *
 * PURPOSE
 *		To draw a section of a room on the current work page.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *	02/22/90 Thursday (JMA) - Changed out of range check to consider the
 *				number of global tiles.
 *
 *
 * SEE ALSO
 *
*/
void DrawPartOfRoom (
	tUMEPackRoomType	*tproom,
	short				 room,
	short				 xoff,
	short				 yoff,
	short				 txoff,
	short				 tyoff,
	short				 twidth,
	short				 theight
)
{
	GShape	**imijz			= tproom->CTI->TileShapes;
	UBYTE	**layers		= tproom->Rooms[room];
	WORD	 lcount			= tproom->NumLayers;
	WORD	 width			= tproom->RoomWidth;
	WORD	 height			= tproom->RoomHeight;
	WORD	 tilewidth		= tproom->TileWidth;
	WORD	 tileheight		= tproom->TileHeight;
	WORD	*tsetoffsets	= tproom->CTI->TileSetOffsets;
	short	 totaltiles		= tproom->CTI->NumTiles + 
								tproom->CTI->NumGlobalTiles;
	short	 i;
	short 	 xpos;
	short	 ypos;
	short	 endxpos	 = twidth * tilewidth;
 	short	 endypos	 = theight * tileheight;
	GShape	*imijptr;
	UBYTE	*HIlayerptr;
	UBYTE	*LOlayerptr;
	UBYTE	 modulo		 = width - twidth;
	WORD	 imijnum;
	WORD	 area		 = width * height;

 	for (i = 0; i < lcount; i++, layers++) {
		HIlayerptr = *layers + txoff + tyoff * width;
		LOlayerptr = HIlayerptr + area;
		for (ypos = 0; ypos != endypos; ypos += tileheight) {
			for (xpos = 0; xpos != endxpos; xpos += tilewidth) {
				if (!(*HIlayerptr)) { /* is empty */
					imijnum = 0;
				} else if (*HIlayerptr & 0x80) { /* is CTI index */
					GET_WORD (imijnum, (*HIlayerptr & 0x7F), *LOlayerptr);
				} else {						 /* is image tile */
					imijnum = tsetoffsets[*HIlayerptr] + *LOlayerptr;
				}

				if (imijnum > 0) {
					if (imijnum > totaltiles) {
						puts ("Bad tile image num\n");
					} else {
						imijptr = imijz[imijnum];
						if (i) {
							DrawGShape (imijptr, xoff + xpos, yoff + ypos);
						} else {
							DrawGShapeNoTrans (imijptr, xoff + xpos, yoff + ypos);
						}
					}
				}
				++HIlayerptr;
				++LOlayerptr;
			}
			HIlayerptr += modulo;
			LOlayerptr += modulo;
		}
	}
} /* DrawPartOfRoom */

